#include <QCoreApplication>
#include <cassert>
#include <iostream>
#include <QJsonArray>
#include <QJsonObject>
#include <QJsonDocument>
#include <QIODevice>
#include <QFile>
#include <QDir>
#include <QTimer>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>

using namespace std;
struct Trade {
    double price;
    string currency;
    string item;
    int amount;
};
vector<Trade> trades;

// Multipliers for each currency, normalized to emeralds
map<string, double> currencyMultipliers;
vector<string> validCurrencies;

QString dataURL = "http://pvc-utils.xyz/resources/data.json";
QString currencyMultipliersURL = "http://pvc-utils.xyz/resources/currencymultipliers.json";


bool fileExists(const QString& filename) {
    QFile file(filename);
    return file.exists();
}

QJsonDocument downloadJson(QUrl URL){
    QNetworkAccessManager* manager = new QNetworkAccessManager();


    QNetworkRequest networkRequest(URL);
    QNetworkReply* reply = manager->get(networkRequest);

    QEventLoop eventLoop;
    QObject::connect(reply, &QNetworkReply::finished, &eventLoop, &QEventLoop::quit);
    eventLoop.exec();

    QByteArray replyData = reply->readAll();
    QJsonDocument jsonDocument = QJsonDocument::fromJson(replyData);
    reply->deleteLater();
    return jsonDocument;
}


double findMedianExchangeRate(const string& currency2, const string& currency1) {
    // Collect all exchange rates from currency1 to currency2
    vector<double> rates;
    for (const Trade& trade : trades) {
        if (trade.currency == currency1 && trade.item == currency2) {
            rates.push_back(trade.price / trade.amount);
        }
    }

    // If there are no trades from currency1 to currency2, return 0
    if (rates.empty()) {
        return 0.0;
    }

    // Sort the rates
    sort(rates.begin(), rates.end());

    // Find the median
    size_t size = rates.size();
    if (size % 2 == 0) {
        // Even number of rates, return the average of the two middle elements
        return (rates[size / 2 - 1] + rates[size / 2]) / 2.0;
    } else {
        // Odd number of rates, return the middle element
        return rates[size / 2];
    }
}


void parse(const QJsonObject& data, vector<Trade>& trades, vector<string> validCurrencies) {
    cout << "Parsing JSON data..." << endl;
    trades.clear();

    if (data.contains("data") && data["data"].isArray()) {
        QJsonArray dataArray = data["data"].toArray();

        for (const QJsonValue& entryValue : dataArray) {
            QJsonObject entry = entryValue.toObject();
            if (entry.contains("recipes") && entry["recipes"].isArray()) {
                QJsonArray recipesArray = entry["recipes"].toArray();
                for (const QJsonValue& recipeValue : recipesArray) {
                    QJsonObject recipe = recipeValue.toObject();

                    QJsonObject item1 = recipe["item1"].toObject();
                    QString currency1 = item1["type"].toString();
                    double price1 = item1["amount"].toDouble();

                    double price2 = 0.0;
                    if (recipe.contains("item2")) {
                        QJsonObject item2 = recipe["item2"].toObject();
                        price2 = item2["amount"].toDouble();
                    }

                    QJsonObject resultItem = recipe["resultItem"].toObject();
                    QString item = resultItem["type"].toString();
                    int amount = resultItem["amount"].toInt();
                    int stock = recipe["stock"].toInt();

                    if ( !item.contains("SHULKER_BOX") && price2 == 0 && find(validCurrencies.begin(), validCurrencies.end(), item.toStdString()) != validCurrencies.end() && find(validCurrencies.begin(), validCurrencies.end(), currency1.toStdString()) != validCurrencies.end() && stock != 0) {
                        Trade tradeInstance;
                        tradeInstance.price = price1;
                        tradeInstance.currency = currency1.toStdString();
                        tradeInstance.item = item.toStdString();
                        tradeInstance.amount = amount;

                        trades.push_back(tradeInstance);
                    }
                }
            }
        }
        cout << "Total currency exchanges: " << trades.size() << endl;
    }
}



double findCurrencyMultiplier(string currency1, string currency2) {
    double exchangeRate = findMedianExchangeRate(currency1, currency2);
    double backwardsRate = findMedianExchangeRate(currency2, currency1);
    return (exchangeRate != 0 && backwardsRate != 0) ? ((exchangeRate + (1 / backwardsRate)) / 2) : (exchangeRate == 0 ? 1 / backwardsRate : exchangeRate);
}

bool saveMapToJson(const std::map<std::string, double>& map, const QString& filename) {
    // Convert the map to a QJsonObject
    QJsonObject jsonObj;
    for (const auto& pair : map) {
        jsonObj[QString::fromStdString(pair.first)] = pair.second;
    }

    // Create a QJsonDocument from the QJsonObject
    QByteArray jsonDoc = QJsonDocument(jsonObj).toJson(QJsonDocument::Indented);

    // Save the QJsonDocument to a file
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly)) {
        qWarning() << "Couldn't open file" << filename << "for writing.";
        return false;
    }

    // Write the JSON document to the file
    file.write(jsonDoc);
    file.close();
    qDebug() << QString(jsonDoc);

    return true;
}



int main(int argc, char* argv[]) {
    QCoreApplication app(argc, argv);

    QTimer::singleShot(0, [](){
        // Fetch currency multipliers
        QJsonObject currencyMultipliersJson = downloadJson(currencyMultipliersURL).object();
        map<string, double> currencyMultipliers;

        for (auto it = currencyMultipliersJson.begin(); it != currencyMultipliersJson.end(); ++it) {
            QString key = it.key();
            QJsonValue value = it.value();

            // Ensure that the value is a double before inserting
            if (value.isDouble()) {
                currencyMultipliers.insert({key.toStdString(), value.toDouble()});
                validCurrencies.emplace_back(key.toStdString());
            }
        }

        auto dataJson = downloadJson(dataURL).object();
        assert (! dataJson.isEmpty());
        parse(dataJson, trades, validCurrencies);
        cout << "Success downloading and parsing data!" << endl;

        double oldDiamondMultiplier = currencyMultipliers["DIAMOND"];

        double diamondMultiplier = findCurrencyMultiplier("DIAMOND", "EMERALD_BLOCK");
        double netheriteMultiplier = findMedianExchangeRate("NETHERITE_INGOT", "DIAMOND_BLOCK");
        double ironMultiplier = findCurrencyMultiplier("IRON_BLOCK", "EMERALD_BLOCK");
        double goldMultiplier = findCurrencyMultiplier("GOLD_BLOCK", "EMERALD_BLOCK");
        cout << "Diamond: " << diamondMultiplier << ", Netherite: " << netheriteMultiplier << ", Iron: " << ironMultiplier << ", Gold: " << goldMultiplier << endl;

        if (diamondMultiplier) {
            currencyMultipliers["DIAMOND"] = diamondMultiplier;
            currencyMultipliers["DIAMOND_BLOCK"] = diamondMultiplier * 9;

            if (netheriteMultiplier) {
                currencyMultipliers["ANCIENT_DEBRIS"] = 9 * diamondMultiplier * netheriteMultiplier / 4;
                currencyMultipliers["NETHERITE_SCRAP"] =  9 * diamondMultiplier * netheriteMultiplier / 4;
                currencyMultipliers["NETHERITE_INGOT"] =  9 * diamondMultiplier * netheriteMultiplier;
                currencyMultipliers["WITHER_SKELETON_SKULL"] =  9 * diamondMultiplier * netheriteMultiplier;
            } else {
                currencyMultipliers["ANCIENT_DEBRIS"] = currencyMultipliers["ANCIENT_DEBRIS"] * (diamondMultiplier / oldDiamondMultiplier);
                currencyMultipliers["NETHERITE_SCRAP"] = currencyMultipliers["NETHERITE_SCRAP"] * (diamondMultiplier / oldDiamondMultiplier);
                currencyMultipliers["NETHERITE_INGOT"] = currencyMultipliers["NETHERITE_INGOT"] * (diamondMultiplier / oldDiamondMultiplier);
                currencyMultipliers["WITHER_SKELETON_SKULL"] = currencyMultipliers["WITHER_SKELETON_SKULL"] * (diamondMultiplier / oldDiamondMultiplier);
            }
        }

        if (goldMultiplier) {
            currencyMultipliers["GOLD_INGOT"] = goldMultiplier / 9;
            currencyMultipliers["GOLD_BLOCK"] = goldMultiplier;
        }
        if (ironMultiplier) {
            currencyMultipliers["IRON_INGOT"] = ironMultiplier / 9;
            currencyMultipliers["IRON_BLOCK"] = ironMultiplier;
        }

        saveMapToJson(currencyMultipliers, "resources/currencymultipliers.json");
        QCoreApplication::exit(0);
    });

    return app.exec();
}
